#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/device.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/cdev.h>
#include <linux/slab.h>
#include <asm/uaccess.h>
#include <linux/usb.h>
#include <linux/string.h>

#include <linux/uaccess.h>
#include <linux/compat.h>
#include "rds-01-public.h"
#include "rds-01.h"
#include "touchmdef.h"
#include "rds-01.priv.h" 

typedef struct _ibutton_info
{
  u32 Family;
  u32 Size;
  char Name[32];
}IBUTTON_INFO,*PIBUTTON_INFO;

static const u8 BitList[]=
{
  0x01,
  0x02,
  0x04,
  0x08,
  0x10,
  0x20,
  0x40,
  0x80,
};



IBUTTON_INFO DevList[]=
{
  {0x01,  0,       "DS1990A"  },
  {0x02,  512/8,   "DS1991"   },
  {0x08,  1024/8,  "DS1992"   },
  {0x06,  4096/8,  "DS1993"   },
  {0x04,  4096/8,  "DS1994"   },
  {0x0A,  16384/8, "DS1995"   },
  {0x0C,  65536/8, "DS1996"   },
  
  {0x09,  1024/8,  "DS1982"   },
  {0x0B,  16384/8, "DS1985"   },
  {0x0F,  65536/8, "DS1986"   },
  {0x10,  2,       "DS1920"   },
  
  {0,     0,       ""         },
};





int read_1w(PRDS_01_CTX pDevCtx,u8 *pBuffer,u32 DataSize)
{
  u32 i,j,TxSize,ByteIdx;
  int ret;
  if(DataSize>RDS01_MAX_BLOCK_SIZE)
    return -EFAULT;
  TxSize=DataSize*8;
  ByteIdx=0;
  memset(pBuffer,0,DataSize);
  memset(pDevCtx->pTXWrkBuffer,0xFF,TxSize);
  ret=pl2303_write(pDevCtx,pDevCtx->pTXWrkBuffer,TxSize);
  if(ret<0)
  {
    DebugPrint(KERN_INFO "write ll failed %d\n",ret);
    return ret;  
  }
  mdelay(10);
  ret=pl2303_read(pDevCtx,pDevCtx->pRXWrkBuffer,TxSize*2);
  if(ret<0)
  {
    DebugPrint(KERN_INFO "read ll failed %d\n",ret);
    return ret;
  }
  if(ret<TxSize)
  {
    ret=pl2303_read(pDevCtx,&pDevCtx->pRXWrkBuffer[ret],TxSize-ret);
  }
  ByteIdx=0;
  for(i=0;i<DataSize;i++)
  {
    for(j=0;j<8;j++)
    {
      if(pDevCtx->pRXWrkBuffer[ByteIdx]==0xFF)
        pBuffer[i]|=BitList[j];
      ByteIdx++;
    }
  }
  return ret/8;
}




static int rds01_IsSlPresent(PRDS_01_CTX pDevCtx)
{
  int BytesRead;
  pl2303_set_speed(pDevCtx,9600);
  mdelay(20);
  pDevCtx->pDMACtrlBuf[0]=0xF0;
  pl2303_write(pDevCtx,pDevCtx->pDMACtrlBuf,1);
  mdelay(20);
  BytesRead=pl2303_read(pDevCtx,pDevCtx->pTmpRXBuf,1);
  mdelay(20);
  pl2303_set_speed(pDevCtx,115200);
  //pl2303_set_speed(pDevCtx,230400);
#if 1
  DebugPrint(KERN_INFO "read %d bytes= %x %x\n",BytesRead,pDevCtx->pTmpRXBuf[0],pDevCtx->pTmpRXBuf[1]);
#endif  
  if((pDevCtx->pTmpRXBuf[0]>0x10)&&(pDevCtx->pTmpRXBuf[0]<0xE0))
  {
    //mdelay(10);
    return ETM_OK;
  }
  return -ETM_NO_SLAVE;
}

static int rds01_SkipRom(PRDS_01_CTX pDevCtx)
{
  pDevCtx->pDMACtrlBuf[0]=CMD_SKIP_ROM;
  write_1w(pDevCtx,pDevCtx->pDMACtrlBuf,1);
  return 0;
}

static int rds01_ResetConn(PRDS_01_CTX pDevCtx)
{
  rds01_IsSlPresent(pDevCtx);
  mdelay(20);
  rds01_SkipRom(pDevCtx);
  return 0;
}


static int rds01_ReadRom(PRDS_01_CTX pDevCtx,u8  __user *pValue)
{
  int ret;
  u8 Crc;
//  DebugPrint(KERN_INFO "rds01_ReadRom\n");

  pDevCtx->pDMACtrlBuf[0]=CMD_READ_ROM;
  ret=write_1w(pDevCtx,pDevCtx->pDMACtrlBuf,1);
  if(ret<0)
  {
    DebugPrint(KERN_INFO "write 1w failed %d\n",ret);
    return ret;
  }
  mdelay(20);
  ret=read_1w(pDevCtx,pDevCtx->pTmpRXBuf,ROM_SIZE);
  if(ret<0)
  {
    DebugPrint(KERN_INFO "read 1w failed %d\n",ret);
    return ret;
  }
  if(__copy_to_user(pValue,pDevCtx->pTmpRXBuf,ROM_SIZE))
  {
    return -EFAULT;
  }
  //rds01_ResetConn(pDevCtx);
  Crc=CalcTMCrc(pDevCtx->pTmpRXBuf,ROM_SIZE);
  if(Crc)
  {
    DebugPrint(KERN_INFO "crc error\n");
    return -ETM_BAD_CRC;
  } 
  return ETM_OK;
}


#define READ_BLK_SIZE (TM_SCRATCHPAD_SIZE)

static int rds01_MemRead(PRDS_01_CTX pDevCtx,u32 Size,u8 __user * pData)
{
  int ret,i,tmp,retG;
  u32 TmpOfft;
  if(pDevCtx->IsRemoved)
    return -ENODEV;
  rds01_ResetConn(pDevCtx);
#if 0  
  DebugPrint(KERN_INFO "rds01_MemRead\n");
#endif  
  pDevCtx->pDMACtrlBuf[0]=CMD_READ_MEM;
  pDevCtx->pDMACtrlBuf[1]=(u8)(pDevCtx->CurOffset&0xFF);
  pDevCtx->pDMACtrlBuf[2]=(u8)((pDevCtx->CurOffset>>8)&0xFF);
  ret=write_1w(pDevCtx,pDevCtx->pDMACtrlBuf,3);
  if(ret<0)
  {
    DebugPrint(KERN_INFO "write 1w failed %d\n",ret);
    retG=ret;
    goto cleanup;
  }
  mdelay(20);
  TmpOfft=0;
  for(i=0;i<Size/READ_BLK_SIZE;i++)
  {
    if((tmp=read_1wEx(pDevCtx,pDevCtx->pRXBlockBuffer,READ_BLK_SIZE))<0)
    {
      retG=tmp;
      goto cleanup;
    }
    if(__copy_to_user(&pData[TmpOfft],pDevCtx->pRXBlockBuffer,READ_BLK_SIZE))
    {
      retG=-EFAULT;
      goto cleanup;
    } 
    TmpOfft+=READ_BLK_SIZE;
    pDevCtx->CurOffset+=READ_BLK_SIZE;
  }
  if((tmp=read_1wEx(pDevCtx,pDevCtx->pRXBlockBuffer,Size%READ_BLK_SIZE))<0)
  {
    retG=tmp;
    goto cleanup;
  }
  if(__copy_to_user(&pData[TmpOfft],pDevCtx->pRXBlockBuffer,Size%READ_BLK_SIZE))
  {
    retG=-EFAULT;
    goto cleanup;
  }
  pDevCtx->CurOffset+=Size%READ_BLK_SIZE;
  retG=Size;
cleanup:
  rds01_IsSlPresent(pDevCtx);
  return Size;
}


static int rds01_WriteBlock(PRDS_01_CTX pDevCtx,u32 Offset,u32 Size,u8* pData)
{
  u8 CmdBuffer[8];
  u8 RespBuffer[8];  
  if(Size>TM_SCRATCHPAD_SIZE)
    return -EFAULT;
  if(pDevCtx->IsRemoved)
    return -ENODEV;
  rds01_ResetConn(pDevCtx);
#if 1  
  DebugPrint(KERN_INFO "write %d bytes offset %d\n",Size,Offset);
#endif  
  pDevCtx->pTXBlockBuffer[0]=CMD_WRITE_SCRATCHPAD;
  pDevCtx->pTXBlockBuffer[1]=(u8)(Offset&0xFF);
  pDevCtx->pTXBlockBuffer[2]=(u8)((Offset>>8)&0xFF);
  memcpy(&pDevCtx->pTXBlockBuffer[3],pData,Size);
  write_1wEx(pDevCtx,pDevCtx->pTXBlockBuffer,Size+3);
  rds01_ResetConn(pDevCtx);
  CmdBuffer[0]=CMD_READ_SCRATCHPAD;
  write_1w(pDevCtx,CmdBuffer,1);
  read_1w(pDevCtx,&pDevCtx->pRXBlockBuffer[0],3);
#if 1  
  DebugPrint(KERN_INFO "status %x\n",pDevCtx->pRXBlockBuffer[2]);
#endif  
  if(pDevCtx->pRXBlockBuffer[2]&0x60)
  {
    DebugPrint(KERN_INFO "invalid status %x\n",pDevCtx->pRXBlockBuffer[2]);
    return -ETM_ERR_WRITE;
  }
  read_1wEx(pDevCtx,&pDevCtx->pRXBlockBuffer[3],Size);
  if(memcmp(&pDevCtx->pRXBlockBuffer[3],pData,Size)!=0)
  {
    DebugPrint(KERN_INFO "data mismatch\n");
    return -ETM_ERR_WRITE;
  }
  rds01_ResetConn(pDevCtx);
  pDevCtx->pDMACtrlBuf[0]=CMD_COPY_SCRATCHPAD;
  pDevCtx->pDMACtrlBuf[1]=pDevCtx->pRXBlockBuffer[0];
  pDevCtx->pDMACtrlBuf[2]=pDevCtx->pRXBlockBuffer[1];
  pDevCtx->pDMACtrlBuf[3]=pDevCtx->pRXBlockBuffer[2];
  write_1w(pDevCtx,pDevCtx->pDMACtrlBuf,4);
  mdelay(20);
  read_1w(pDevCtx,&RespBuffer[0],4);
#if 1  
  DebugPrint(KERN_INFO "status %x\n",RespBuffer[0]);
#endif  
  return Size;
}


int priv_fops_open(struct inode *inode, struct file *filp)
{
  unsigned int minor = iminor(inode);
  PRDS_01_CTX pDevCtx;
  if((pDevCtx=PrivFindDevCtx(minor))==NULL)
  {
    DebugPrint(KERN_ERR "rds-01: can not find dev %d\n",minor);
    return -ENXIO;
  }
  if(pDevCtx->IsOpen)
  {
    DebugPrint(KERN_ERR "rds-01: dev %d already opened\n",minor);
    return -EBUSY;
  }
  pl2303_open(pDevCtx);
  kref_get(&pDevCtx->kref);
  filp->private_data=pDevCtx;
  pDevCtx->IsOpen=1;
  pDevCtx->CurOffset=0;
  return 0;
}

int priv_fops_release(struct inode *inode, struct file *filp)
{
  PRDS_01_CTX pDevCtx;
  if((pDevCtx=(PRDS_01_CTX)filp->private_data)==NULL)
  {
    DebugPrint(KERN_ERR "rds-01:file priv data NULL!!\n");
    return -ENXIO;
  }
  pDevCtx->IsOpen=0;
  pDevCtx->CurOffset=0;
  filp->private_data=NULL; 
  kref_put(&pDevCtx->kref,priv_dev_delete);
  //DebugPrint(KERN_INFO "release ok\n");
  return 0;
}

long priv_fops_ioctl(struct file *filp,unsigned int cmd,unsigned long arg)
{
  int res;
  PRDS_01_CTX pDevCtx;
  if((pDevCtx=(PRDS_01_CTX)filp->private_data)==NULL)
  {
    DebugPrint(KERN_ERR "rds-01: file priv data NULL!!\n");
    return -ENXIO;
  }
  switch(cmd)
  {
    case IOCTL_TM_CHK_SL:
      return rds01_IsSlPresent(pDevCtx);
    case IOCTL_TM_GET_ROM:
      res=rds01_ReadRom(pDevCtx,(u8 __user*)arg);
      rds01_ResetConn(pDevCtx);
      return res;
    case IOCTL_TM_SKIP_ROM:
      return rds01_SkipRom(pDevCtx);
  }
  return -EIO;  
}

ssize_t priv_fops_read(struct file *filp, char __user *buf, size_t count, loff_t *f_pos)
{
  PRDS_01_CTX pDevCtx;
  if((pDevCtx=(PRDS_01_CTX)filp->private_data)==NULL)
  {
    DebugPrint(KERN_ERR "rds-01: file priv data NULL!!\n");
    return -ENXIO;
  }
  return rds01_MemRead(pDevCtx,count,buf);
}

ssize_t priv_fops_write(struct file *filp, const char __user *buf,
                            size_t count, loff_t *f_pos)
{
  u32 TmpOffset,BlkSize,i;
  u32 WriteSize;
  int ret;
  PRDS_01_CTX pDevCtx;
  if((pDevCtx=(PRDS_01_CTX)filp->private_data)==NULL)
  {
    DebugPrint(KERN_ERR "rds-01: file priv data NULL!!\n");
    return -ENXIO;
  }
  WriteSize=0;
  TmpOffset=0;
  if(pDevCtx->CurOffset%TM_SCRATCHPAD_SIZE)
  {
    BlkSize=TM_SCRATCHPAD_SIZE-(pDevCtx->CurOffset%TM_SCRATCHPAD_SIZE);
    if(BlkSize>count)
      BlkSize=count;
    if(__copy_from_user(pDevCtx->pWriteBuffer,buf,BlkSize))
      return -EFAULT;
    if((ret=rds01_WriteBlock(pDevCtx,pDevCtx->CurOffset,BlkSize,pDevCtx->pWriteBuffer))<0)
      return ret;
    pDevCtx->CurOffset+=BlkSize;
    WriteSize+=BlkSize;
    count-=BlkSize;
  }
  if(count==0)
    return WriteSize;
  for(i=0;i<count/TM_SCRATCHPAD_SIZE;i++)
  {
    if(__copy_from_user(pDevCtx->pWriteBuffer,&buf[WriteSize],TM_SCRATCHPAD_SIZE))
      return -EFAULT;
    if((ret=rds01_WriteBlock(pDevCtx,pDevCtx->CurOffset,TM_SCRATCHPAD_SIZE,pDevCtx->pWriteBuffer))<0)
      return ret;
    pDevCtx->CurOffset+=TM_SCRATCHPAD_SIZE;
    WriteSize+=TM_SCRATCHPAD_SIZE;
  }
  if(count%TM_SCRATCHPAD_SIZE)
  {
    if(__copy_from_user(pDevCtx->pWriteBuffer,&buf[WriteSize],count%TM_SCRATCHPAD_SIZE))
      return -EFAULT;
    if((ret=rds01_WriteBlock(pDevCtx,pDevCtx->CurOffset,count%TM_SCRATCHPAD_SIZE,pDevCtx->pWriteBuffer))<0)
      return ret;
    pDevCtx->CurOffset+=count%TM_SCRATCHPAD_SIZE;
    WriteSize+=count%TM_SCRATCHPAD_SIZE;
  }
  return count;
}

loff_t priv_fops_lseek(struct file *filp, loff_t Offset,int whence)
{
  PRDS_01_CTX pDevCtx;
  if((pDevCtx=(PRDS_01_CTX)filp->private_data)==NULL)
  {
    DebugPrint(KERN_ERR "rds-01: file priv data NULL!!\n");
    return -ENXIO;
  }
  switch(whence)
  {
    case SEEK_SET:
      pDevCtx->CurOffset=Offset;
      break;
    case SEEK_CUR:
      pDevCtx->CurOffset+=Offset;
      break;
    default:
      return -ESPIPE;
  }
  DebugPrint(KERN_INFO "new offset %d\n",pDevCtx->CurOffset);
  return pDevCtx->CurOffset;
}
